"""Example how to to simply send RF commands.

A running pilight-daemon is needed.
"""

from pilight import pilight

# pylint: disable=C0103
if __name__ == '__main__':
    # Create new pilight connection that runs on localhost with port 5000
    pilight_client = pilight.Client(host='127.0.0.1', port=5000)

    # Send a good code. It is checked to be acknoledged by the piligt-daemon.
    # Data from https://manual.pilight.org/en/api
    pilight_client.send_code(data={"protocol": ["kaku_switch"],
                                   "id": 1,
                                   "unit": 0,
                                   "off": 1})

    # Send a wrong code that lead to an IO error since the pilight-daemon
    # rejects it. Data from https://manual.pilight.org/en/api
    pilight_client.send_code(data={"protocol": ["kaku_switch"],
                                   "id": 0,  # ID has to be > 0
                                   "unit": 0,
                                   "off": 1})



# http://stackoverflow.com/questions/17583443/what-is-the-correct-way-to-share-package-version-with-setup-py-and-the-package
from pkg_resources import get_distribution, DistributionNotFound

__project__ = 'pilight'
__version__ = None  # required for initial installation

try:
    __version__ = get_distribution(__project__).version
except DistributionNotFound:
    VERSION = __project__ + '-' + '(local)'
else:
    VERSION = __project__ + '-' + __version__


"""Example how to use the pilight python client asynchronously.

A running pilight-daemon is needed, otherwise nothing will be shown here.
"""

import time
from pilight import pilight


def handle_code(code):  # Simply print received data from pilight
    """Handle to just prints the received code."""
    print(code)

# pylint: disable=C0103
if __name__ == '__main__':
    # Suscribe receiver to core info to have something to receive
    # asynchrounosly. With std. settings the receiver calls the
    # data handle only on received codes!
    recv_ident = {
        "action": "identify",
        "options": {
            "core": 1,  # To get CPU load and RAM of pilight daemon
            "receiver": 1  # To receive the RF data received by pilight
        }
    }

    # Create new pilight connection that runs on localhost with port 5000
    # Turn off to print core info and send info of the pilight-daemon
    pilight_client = pilight.Client(host='127.0.0.1',
                                    port=5000,
                                    recv_ident=recv_ident,
                                    recv_codes_only=False)

    # Set a data handle that is called on received data
    pilight_client.set_callback(handle_code)
    pilight_client.start()  # Start the receiver

    # While the received data is handled you can also send commands
    # Data from https://manual.pilight.org/en/api
    data = {"protocol": ["kaku_switch"],
            "id": 1,
            "unit": 0,
            "off": 1}

    for i in range(5):
        data['id'] = i + 1  # ID > 0 for the kaku_switch protocoll
        time.sleep(2)
        # Will create a sender message received and printed by the receiver
        pilight_client.send_code(data)



""" This is a helper class providing a pilight-daemon in python.

    This daemon just simulates the behavior of the real pilight-daemon
    and is used for testing only. Of cause only a few important
    commands are supported.

    This is a very hackish synchronous daemon that is
    not an example for good socket servers!
"""

import datetime
import json
import socket
import threading
import time
import sys

if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue

# Settings for the pilight-daemon simulation
HOST = '127.0.0.1'
PORT = 5000

SEND_DELAY = 0.1  # How often a fake code is send in seconds

FAKE_DATA = {"origin": "receiver",  # Fake data being send
             "repeats": 1,
             "message": {
                 "protocol": ["kaku_switch"],
                 "id": 0,
                 "unit": 0,
                 "off": 1}}


class PilightDaemon(object):

    """Provide a pilight-daemon in with-statement."""

    def __init__(self, host=HOST, port=PORT, send_codes=False):
        self.host = host
        self.port = port
        self.send_codes = send_codes
        self.pilight_daemon = PilightDeamonSim(
            self.host, self.port, self.send_codes)

    def __enter__(self):
        self.pilight_daemon.start()
        time.sleep(0.1)
        return self.pilight_daemon

    def __exit__(self, _, value, traceback):
        self.pilight_daemon.stop()
        time.sleep(0.1)


class PilightDeamonSim(threading.Thread):

    """Simulate the pilight-daemon API for testing."""

    def __init__(self, host, port, send_codes):
        self.send_codes = send_codes

        # Setup thread
        threading.Thread.__init__(self)
        self.daemon = True
        self._stop_thread = threading.Event()
        self._lock = threading.Lock()
        self._data = queue.Queue()

        # Setup server socket
        self.server_socket = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)

        time.sleep(0.1)
        self.server_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        time.sleep(0.1)
        self.server_socket.settimeout(0.01)  # Unset blocking

        # Try to bin to address. Maybe not available yet thus
        # try up to 10 times waiting up to 10 seconds.
        # Idea from http://stackoverflow.com/questions/6380057/
        # python-binding-socket-address-already-in-use
        for _ in range(10):
            try:
                self.server_socket.bind((host, port))
                break
            except socket.error:
                time.sleep(1)
        else:  # Called when for loop not breaked
            raise RuntimeError('Cannot create socket connection')

        self.server_socket.listen(2)  # Allow 2 connections
        self.client_sockets = []

        self.last_send = datetime.datetime.now()

    def run(self):
        """Simple infinite loop handling socket connections."""
        with self._lock:
            while not self._stop_thread.wait(0.01):
                self._handle_client_connections()
                self._handle_client_data()
                self._send_codes()

        # Close client connections
        for client_socket in self.client_sockets:
            client_socket.shutdown(socket.SHUT_RDWR)
            client_socket.close()

    def _send_codes(self):
        if self.send_codes:
            if (((datetime.datetime.now() -
                  self.last_send).total_seconds() > SEND_DELAY)):
                if len(self.client_sockets) > 0:
                    for i in range(10):  # Send data 10 times to simulate button press
                        fake_data = FAKE_DATA.copy()
                        fake_data['repeats'] = i + 1
                        self.client_sockets[0].send(
                            json.dumps(fake_data).encode())
                        time.sleep(0.01)

    def _handle_client_connections(self):
        def _new_client(client_socket):
            """Handle new client connection.

            Fake pilight-daemon protocol by returning success on identifiy action.
            """
            def _acknowledge_connection(messages):
                for message in messages:  # Loop over received messages
                    if message:  # Can be empty due to splitlines
                        message_dict = json.loads(message.decode())
                        if "action" in message_dict:
                            if message_dict["action"] == "identify":
                                client_socket.sendall(json.dumps({'status': 'success'}).encode())
                            else:
                                client_socket.sendall(json.dumps({'status': 'failure'}).encode())
                            break

            client_socket.settimeout(0.01)  # Unset blocking
            messages = client_socket.recv(1024).splitlines()
            _acknowledge_connection(messages)
            return client_socket
        try:
            self.client_sockets.append(
                _new_client(self.server_socket.accept()[0]))
        except socket.timeout:
            pass

    def _handle_client_data(self):
        def _handle_message(client_socket):
            """Called in poll loop to handle messages."""
            try:
                messages = client_socket.recv(1024).splitlines()
                for message in messages:  # Loop over received messages
                    if message:  # Can be empty due to splitlines
                        message_dict = json.loads(message.decode())
                        self._data.put(message_dict)
                        if message_dict["code"]["protocol"] == "daycom":
                            client_socket.sendall(
                                json.dumps({'status': 'success'}).encode())
                        else:
                            client_socket.sendall(
                                json.dumps({'status': 'failure'}).encode())
            except socket.error:
                pass

        for client_socket in self.client_sockets:  # Simple poll for data
            try:
                _handle_message(client_socket)
            except socket.timeout:  # We poll, thus timeout for no data
                pass

    def get_data(self):
        """Return received data."""
        return self._data.get()

    def stop(self):
        """Called to stop the reveiver thread."""
        self._stop_thread.set()
        with self._lock:  # Receive thread might use the socket
            try:
                self.server_socket.shutdown(socket.SHUT_RDWR)
            except socket.error:  # Connection already shutdown
                pass
            self.server_socket.close()



"""This module implements a client to interface the pilight-daemon.

More information about pilight is here: https://www.pilight.org/.
"""

import threading
import socket
import json
import logging


class Client(threading.Thread):

    """This client interfaces with the pilight-daemon (https://www.pilight.org/).

    Sending and receiving codes is implemented in an asychronous way.
    A callback function can be defined that reacts on received data.

    All pilight-send commands can be used by this client. Documentation
    can be found here https://wiki.pilight.org/doku.php/psend.
    Also check https://manual.pilight.org/en/api.

    :param host: Address where the pilight-daemon intance runs
    :param port: Port of the pilight-daemon on the host
    :param timeout: Time until a time out exception is raised when connecting
    :param recv_ident: The identification of the receiver to sucribe
    to the pilight-daemon topics (https://manual.pilight.org/en/api)
    :param recv_codes_only: If True: only call the callback function when the
    pilight-daemon received a code, not for status messages etc.
    :param veto_repeats: If True: only call the callback function when the
    pilight-daemon received a new code, not the same code repeated.
    Repeated codes happen quickly when a button is pressed.
    """

    # pylint: disable=too-many-arguments, too-many-instance-attributes

    def __init__(self, host='127.0.0.1', port=5000, timeout=1,
                 recv_ident=None, recv_codes_only=True, veto_repeats=True):
        """Initialize the pilight client.

        The readout thread is not started automatically.
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self._stop_thread = threading.Event()
        self._lock = threading.Lock()
        self.recv_codes_only = recv_codes_only
        self.veto_repeats = veto_repeats

        # Identify client (https://manual.pilight.org/en/api)
        client_identification_sender = {
            "action": "identify",
            "options": {
                # To get CPU load and RAM of pilight daemon, is neverless
                # ignored by daemon ...
                "core": 0,
                "receiver": 0,  # To receive the RF data received by pilight
                "config": 0
            }
        }

        if recv_ident:
            client_identification_receiver = recv_ident
        else:
            client_identification_receiver = {
                "action": "identify",
                "options": {
                    "core": 0,  # To get CPU load and RAM of pilight daemon
                    # To receive the RF data received by pilight
                    "receiver": 1,
                    "config": 0,
                    "forward": 0
                }
            }

        # Open 2 socket connections, one for sending one for receiving data
        # That is the simplest approach to allow asynchronus communication with
        # the pilight daemon
        self.receive_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.send_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Timeout to allow receiver thread termination and to restrict blocking
        # connection time
        self.receive_socket.settimeout(timeout)
        self.send_socket.settimeout(timeout)

        self.receive_socket.connect((host, port))
        self.send_socket.connect((host, port))

        # Identify this clients sockets at the pilight-deamon
        self.receive_socket.send(
            json.dumps(client_identification_receiver).encode())
        answer_1 = json.loads(self.receive_socket.recv(1024).decode())
        self.send_socket.send(
            json.dumps(client_identification_sender).encode())
        answer_2 = json.loads(self.send_socket.recv(1024).decode())

        # Check connections are acknowledged
        if ('success' not in answer_1['status'] or
                'success' not in answer_2['status']):
            raise IOError(
                'Connection to the pilight daemon failed. Reply %s, %s',
                answer_1, answer_2)

        self.callback = None

    def set_callback(self, function):
        """Function to be called when data is received."""
        self.callback = function

    def stop(self):
        """Called to stop the reveiver thread."""
        self._stop_thread.set()
        # f you want to close the connection in a timely fashion,
        # call shutdown() before close().
        with self._lock:  # Receive thread might use the socket
            self.receive_socket.shutdown(socket.SHUT_RDWR)
            self.receive_socket.close()

        self.send_socket.shutdown(socket.SHUT_RDWR)
        self.send_socket.close()

    def run(self):  # Thread for receiving data from pilight
        """Receiver thread function called on Client.start()."""
        logging.debug('Pilight receiver thread started')
        if not self.callback:
            raise RuntimeError('No callback function set, cancel readout thread')

        def handle_messages(messages):
            """Call callback on each receive message."""
            for message in messages:  # Loop over received  messages
                if message:  # Can be empty due to splitlines
                    message_dict = json.loads(message.decode())
                    if self.recv_codes_only:
                        # Filter: Only use receiver messages
                        if 'receiver' in message_dict['origin']:
                            if self.veto_repeats:
                                if message_dict['repeats'] == 1:
                                    self.callback(message_dict)
                            else:
                                self.callback(message_dict)
                    else:
                        self.callback(message_dict)

        while not self._stop_thread.isSet():
            try:  # Read socket in a non blocking call and interpret data
                # Sometimes more than one JSON object is in the stream thus
                # split at \n
                with self._lock:
                    messages = self.receive_socket.recv(1024).splitlines()
                handle_messages(messages)
            except (socket.timeout, ValueError):  # No data
                pass
        logging.debug('Pilight receiver thread stopped')

    def send_code(self, data, acknowledge=True):
        """Send a RF code known to the pilight-daemon.

        For protocols look at https://manual.pilight.org/en/api.
        When acknowledge is set, it is checked if the code was issued.
        :param data: Dictionary with the data
        :param acknowledge: Raise IO exception if the code is not
        send by the pilight-deamon
        """
        if "protocol" not in data:
            raise ValueError(
                'Pilight data to send does not contain a protocol info. '
                'Check the pilight-send doku!', str(data))

        # Create message to send
        message = {
            "action": "send",  # Tell pilight daemon to send the data
            "code": data,
        }

        # If connection is closed IOError is raised
        self.send_socket.sendall(json.dumps(message).encode())

        if acknowledge:  # Check if command is acknowledged by pilight daemon
            messages = self.send_socket.recv(1024).splitlines()
            received = False
            for message in messages:  # Loop over received messages
                if message:  # Can be empty due to splitlines
                    acknowledge_message = json.loads(message.decode())
                    # Filter correct message
                    if ('status' in acknowledge_message and
                            acknowledge_message['status'] == 'success'):
                        received = True
            if not received:
                raise IOError('Send code failed. Code: %s', str(data))



"""Tests the pilight client.

Connects to a simulation of a pilight-daemon.
"""

import unittest
import time
from mock import patch, call

try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import StringIO  # Python 3

from pilight import pilight
from pilight.test import pilight_daemon


def _callback(_):
    """"Function to be called in unit test."""
    pass


class TestClient(unittest.TestCase):

    """Initialize unit test case."""

    def test_client_connection(self):
        """Test for successfull pilight daemon connection."""
        with pilight_daemon.PilightDaemon():
            pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT)

    def test_client_connection_fail(self):
        """Test for failing pilight daemon connection."""
        with pilight_daemon.PilightDaemon():
            with self.assertRaises(IOError):
                pilight.Client(host='8.8.8.8', port=0)

    def test_send_code(self):
        """Test for successfull code send."""
        with pilight_daemon.PilightDaemon() as my_daemon:
            pilight_client = pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT)
            pilight_client.send_code(data={'protocol': 'daycom'})
            time.sleep(1)

        self.assertEqual(my_daemon.get_data()['code'], {'protocol': 'daycom'})

    def test_send_code_fail(self):
        """Tests for failed code send."""
        with pilight_daemon.PilightDaemon():
            pilight_client = pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT)

            # Test 1: Use unknows protocol
            with self.assertRaises(IOError):
                pilight_client.send_code(data={'protocol': 'unknown'})

            # Test 2: Do not send protocoll info, thus Value Error expected
            with self.assertRaises(ValueError):
                pilight_client.send_code(data={'no_protocol': 'test'})

    def test_api(self):
        """Tests connection with different receiver filter and identification."""
        recv_ident = {
            "action": "identify",
            "options": {
                "core": 1,  # To get CPU load and RAM of pilight daemon
                "receiver": 1  # To receive the RF data received by pilight
            }
        }
        with pilight_daemon.PilightDaemon(send_codes=True):
            pilight_client = pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT,
                                            recv_ident=recv_ident, recv_codes_only=False)
            pilight_client.set_callback(_callback)
            pilight_client.start()
            time.sleep(1)
        pilight_client.stop()

    @patch('pilight.test.test_client._callback')
    def test_receive_code(self, mock):
        """Test for successfull code received."""

        with pilight_daemon.PilightDaemon(send_codes=True):
            pilight_client = pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT)
            pilight_client.set_callback(_callback)
            pilight_client.start()
            time.sleep(2)
        pilight_client.stop()

        # Check if callback was called several times with expected data
        mock.assert_has_calls([call(pilight_daemon.FAKE_DATA)] * 10)

    @patch('pilight.test.test_client._callback')
    def test_no_receive_filter(self, mock):
        """Test for successfull code received."""
        with pilight_daemon.PilightDaemon(send_codes=True):
            pilight_client = pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT, veto_repeats=False)
            pilight_client.set_callback(_callback)
            pilight_client.start()
            time.sleep(2)
        pilight_client.stop()

        # Check if calls with repeats != 1 have been made
        calls = []
        for i in range(10):
            fake_data = pilight_daemon.FAKE_DATA.copy()
            fake_data['repeats'] = i + 1
            calls.append(call(fake_data))

        mock.assert_has_calls(calls)

    def test_invalid_identification(self):
        """Send an invalid identification and check for connection failure."""
        recv_ident = {
            "action": "invalid"
        }

        with self.assertRaises(IOError):
            with pilight_daemon.PilightDaemon(send_codes=True):
                pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT, recv_ident=recv_ident)

    def test_no_callback(self):
        """Test for no callback defined."""
        with pilight_daemon.PilightDaemon(send_codes=True):
            pilight_client = pilight.Client(host=pilight_daemon.HOST, port=pilight_daemon.PORT)
            pilight_client.start()
            time.sleep(1)  # Give time to set thread status

        self.assertFalse(pilight_client.isAlive())
        pilight_client.stop()



"""Example how to handle codes received by the pilight-daemon.

A running pilight-daemon is neededed.
"""

import time
from pilight import pilight


def handle_code(code):  # Simply  print received data from pilight
    """Handle to just prints the received code."""
    print(code)

# pylint: disable=C0103
if __name__ == '__main__':
    # Create new pilight connection that runs on localhost with port 5000
    pilight_client = pilight.Client(host='127.0.0.1', port=5000)

    # Set a data handle that is called on received data
    pilight_client.set_callback(handle_code)
    pilight_client.start()  # Start the receiver

    # You have 10 seconds to print all the data the pilight-daemon receives
    time.sleep(10)
    pilight_client.stop()  # Stop the receiver



#!/usr/bin/env python
from setuptools import setup, find_packages  # This setup relies on setuptools since distutils is insufficient and badly hacked code

version = '0.1.1'
author = 'David-Leon Pohl'
author_email = 'David-Leon.Pohl@rub.de'

setup(
    name='pilight',
    version=version,
    description='A pure python module to connect to a pilight daemon to send and receive commands.',
    url='https://github.com/DavidLP/pilight',
    license='MIT License',
    long_description='',
    author=author,
    maintainer=author,
    author_email=author_email,
    maintainer_email=author_email,
    packages=find_packages(),
    include_package_data=True,  # Accept all data files and directories matched by MANIFEST.in or found in source control
    keywords=['pilight', '433', 'light'],
    platforms='any'
)
